__all__ = [
    "StriveLogger",
    "ConsoleLogger",
    "PythonLogger",
    "UvicornLogger",
]

from strivelogger.logger_implementations.console_logger import ConsoleLogger
from strivelogger.logger_implementations.python_logger import PythonLogger
from strivelogger.logger_implementations.uvicorn_logger import UvicornLogger

from .strive_logger import StriveLogger



import traceback
from typing import Optional

from ..logger_implementation import LoggerImplementation


class ConsoleLogger(LoggerImplementation):
    def __log(self, level: str, message: str, trace_id: Optional[str], extra: dict | None = None) -> None:
        if trace_id:
            message = f"[{trace_id}] {message}"

        level = f"{level}:".ljust(10)
        print(f"{level}{message}")
        if extra:
            print(f"{'EXTRA:'.ljust(10)} {extra}")

    def debug(self, message: str, trace_id: str | None, extra: dict | None = None) -> None:
        self.__log("DEBUG", message, trace_id, extra)

    def info(self, message: str, trace_id: str | None, extra: dict | None = None) -> None:
        self.__log("INFO", message, trace_id, extra)

    def warn(self, message: str, trace_id: str | None, extra: dict | None = None) -> None:
        self.__log("WARN", message, trace_id, extra)

    def error(
        self,
        message: str,
        trace_id: str | None,
        exc_info: BaseException | str | None = None,
        extra: dict | None = None,
    ) -> None:
        self.__log("ERROR", message, trace_id, extra)
        if exc_info:
            if isinstance(exc_info, BaseException):
                exc_info = "\n".join(traceback.format_exception(type(exc_info), exc_info, exc_info.__traceback__))
            print(f"{'EXC_INFO:'.ljust(10)} {exc_info}")



from typing import Optional

from .logger_implementation import LoggerImplementation
from .logger_implementations.null_logger import NullLogger
from .tracer_implementation import TracerImplementation


class StriveLogger:
    _logger: LoggerImplementation = NullLogger()
    _tracer: Optional[TracerImplementation] = None

    @classmethod
    def initialize(
        cls,
        logger: LoggerImplementation,
        tracer: Optional[TracerImplementation] = None,
    ) -> None:
        if logger is not None and isinstance(logger, LoggerImplementation):
            cls._logger = logger

        cls._tracer = tracer

    @classmethod
    def logger_type(cls) -> str:
        return type(cls._logger).__name__

    @classmethod
    def get_trace_id(cls) -> Optional[str]:
        return cls._tracer.get_trace_id() if cls._tracer else None

    @classmethod
    def debug(cls, message: str, extra: dict | None = None) -> None:
        cls._logger.debug(message, cls.get_trace_id(), extra)

    @classmethod
    def info(cls, message: str, extra: dict | None = None) -> None:
        cls._logger.info(message, cls.get_trace_id(), extra)

    @classmethod
    def warn(cls, message: str, extra: dict | None = None) -> None:
        cls._logger.warn(message, cls.get_trace_id(), extra)

    @classmethod
    def error(
        cls,
        message: str,
        exc_info: BaseException | str | None = None,
        extra: dict | None = None,
    ) -> None:
        cls._logger.error(message, cls.get_trace_id(), exc_info, extra)



from ..logger_implementation import LoggerImplementation


class NullLogger(LoggerImplementation):
    """
    This logger is used as the default logger implementation if no other logger is initialized.

    Log messages are discarded.

    There may be valid use-cases for this logger in production, but it's primary purpose to
    avoid null pointer exceptions in the StriveLogger.
    """

    def debug(self, message: str, trace_id: str | None, extra: dict | None = None) -> None:
        pass

    def info(self, message: str, trace_id: str | None, extra: dict | None = None) -> None:
        pass

    def warn(self, message: str, trace_id: str | None, extra: dict | None = None) -> None:
        pass

    def error(
        self,
        message: str,
        trace_id: str | None,
        exc_info: BaseException | str | None = None,
        extra: dict | None = None,
    ) -> None:
        pass



from assertpy import assert_that

from strivelogger.logger_implementations.console_logger import ConsoleLogger
from strivelogger.logger_implementations.null_logger import NullLogger
from strivelogger.strive_logger import StriveLogger

DEFAULT_LOGGER_NAME = type(NullLogger()).__name__


def test_uses_null_logger_by_default():
    assert_that(StriveLogger.logger_type()).is_equal_to(DEFAULT_LOGGER_NAME)


def test_can_log_debug_message_when_uninitialized():
    StriveLogger.debug("test")


def test_can_log_info_message_when_uninitialized():
    StriveLogger.info("test")


def test_can_log_warn_message_when_uninitialized():
    StriveLogger.warn("test")


def test_can_log_error_message_when_uninitialized():
    StriveLogger.error("test")


def test_cannot_unset_logger():
    StriveLogger.initialize(None)
    assert_that(StriveLogger.logger_type()).is_equal_to(DEFAULT_LOGGER_NAME)


def test_cannot_set_invalid_logger():
    class NotALoggingImplementation:
        pass

    StriveLogger.initialize(NotALoggingImplementation())
    assert_that(StriveLogger.logger_type()).is_equal_to(DEFAULT_LOGGER_NAME)


def test_can_set_valid_logger():
    StriveLogger.initialize(ConsoleLogger())
    assert_that(StriveLogger.logger_type()).is_equal_to(type(ConsoleLogger()).__name__)



from abc import ABC, abstractmethod
from typing import Optional


class TracerImplementation(ABC):
    @abstractmethod
    def get_trace_id(self) -> Optional[str]:
        raise NotImplementedError()



from abc import ABC, abstractmethod


class LoggerImplementation(ABC):
    @abstractmethod
    def debug(
        self,
        message: str,
        trace_id: str | None,
        extra: dict | None = None,
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    def info(
        self,
        message: str,
        trace_id: str | None,
        extra: dict | None = None,
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    def warn(
        self,
        message: str,
        trace_id: str | None,
        extra: dict | None = None,
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    def error(
        self,
        message: str,
        trace_id: str | None,
        exc_info: BaseException | str | None = None,
        extra: dict | None = None,
    ) -> None:
        raise NotImplementedError()



import json
import logging
from typing import Any, Dict, Optional, Tuple

from .python_logger import LEVELS, PythonLogger


class UvicornLogger(PythonLogger):
    """
    Use the uvicorn logger instance.

    Equivalent to creating a PythonLogger with logger_name 'uvicorn'.
    """

    def __init__(self, level: LEVELS = "INFO", enable_json: bool = False) -> None:
        super().__init__("uvicorn", level, enable_json)

        if enable_json and "uvicorn.access" in logging.Logger.manager.loggerDict:
            access_logger = logging.getLogger("uvicorn.access")
            if len(access_logger.handlers) > 0:
                access_logger.handlers[0].formatter = UvicornAccessFormatter()


class UvicornAccessFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        log: Dict[str, Any] = {
            "message": record.getMessage(),
        }
        if "HTTP/" in record.msg:
            UvicornAccessFormatter.add_request_info(log, record.args)  # type: ignore

        return json.dumps(log)

    @staticmethod
    def add_request_info(log: Dict[str, Any], args: Optional[Tuple]) -> None:
        if args and len(args) >= 5:
            host = args[0]
            method = args[1]
            path = args[2]
            status_code = args[4]
            is_error_code = status_code >= 400

            log["host"] = host
            log["method"] = method
            log["path"] = path
            log["status_code"] = status_code
            log["success"] = not is_error_code



import json
import logging
import traceback
from typing import Any, Dict, Literal, Optional

from ..logger_implementation import LoggerImplementation

LEVELS = Literal["ERROR", "WARN", "INFO", "DEBUG"]


class PythonLogger(LoggerImplementation):
    """
    Use a python logger instance with the specified name, creating it if necessary.

    If no name is specified the root logger will be used.
    """

    def __init__(
        self,
        logger_name: Optional[str] = None,
        level: LEVELS = "INFO",
        enable_json: bool = False,
    ) -> None:
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(level)
        self.enable_json = enable_json

    def __create_json_record(
        self,
        message: str,
        trace_id: str | None,
        extra: dict | None,
        exc_info: BaseException | str | None,
    ) -> str:
        record: Dict[str, Any] = {
            "message": message,
        }

        if trace_id:
            record["trace_id"] = trace_id
        if extra:
            record["extra"] = extra
        if exc_info:
            if isinstance(exc_info, BaseException):
                record["exc_info"] = traceback.format_exception(exc_info)
            else:
                record["exc_info"] = str(exc_info)

        return json.dumps(record)

    def __log_json(
        self,
        level: int,
        message: str,
        trace_id: str | None,
        extra: dict | None,
        exc_info: BaseException | str | None,
    ) -> None:
        record = self.__create_json_record(message, trace_id, extra, exc_info)
        self.logger.log(level, record)

    def __log_string(
        self,
        level: int,
        message: str,
        trace_id: str | None,
        extra: dict | None,
        exc_info: BaseException | str | None,
    ) -> None:
        if trace_id:
            message = f"[{trace_id}] {message}"

        if extra:
            message = f"{message}\n\t\t{extra}"

        self.logger.log(level, message, exc_info=exc_info)  # type: ignore

    def __log(
        self,
        level: int,
        message: str,
        trace_id: str | None,
        extra: dict | None,
        exc_info: BaseException | str | None = None,
    ) -> None:
        log = self.__log_json if self.enable_json else self.__log_string
        log(
            level,
            message,
            trace_id,
            extra,
            exc_info,
        )

    def debug(self, message: str, trace_id: str | None, extra: dict | None = None) -> None:
        self.__log(logging.DEBUG, message, trace_id, extra)

    def info(self, message: str, trace_id: str | None, extra: dict | None = None) -> None:
        self.__log(logging.INFO, message, trace_id, extra)

    def warn(self, message: str, trace_id: str | None, extra: dict | None = None) -> None:
        self.__log(logging.WARN, message, trace_id, extra)

    def error(
        self,
        message: str,
        trace_id: str | None,
        exc_info: BaseException | str | None = None,
        extra: dict | None = None,
    ) -> None:
        self.__log(logging.ERROR, message, trace_id, extra, exc_info)
